/* 
 * File:   DAQChannelMapConstants.h
 * Author: denis
 *
 * Created on February 3, 2012, 1:46 PM
 */

#ifndef _DAQCHANNELMAPCONSTANTS_H
#define	_DAQCHANNELMAPCONSTANTS_H
#include <stdint.h>

namespace daqchannelmap {      //!< DAQ Channel Map Package

  //<<Below I define the types to be used specifically in our new version of DAQChannelMap.  Contemplate moving into own header file called DAQChannelMapConstants.h...
  
  /// The following unsigned ints represent the fields of bits in which we encode DAQ format information (dchan) and logical (physical) information (lchan).
  /// dchan information has the following constituents: detector, diblock_number, dcm_number, feb_number(or dcm_link as it is defined 
  /// lchan information has the following constituents: detector, plane, cell.
  typedef uint32_t dchan;
  typedef uint32_t lchan;

  typedef uint32_t cell_t;                     // physical cell in module (replacing offline_cell_t)
  typedef uint32_t fiber_t;                    // fiber input to FEB (replacing online_cell_t)
  
  /// Type for plane.
  typedef uint32_t plane_t;                    // global plane (replacing offline_plane_t) 
  
  /// Type for diblocks and blocks.
  typedef uint32_t diblock_t;                  // Diblock number.
  typedef uint32_t block_t;                    // Block number
  
  /// Type for DCM number, counts from 1.
  typedef uint32_t dcm_id_t;                   // DCM number
  
  /// Type for DCM link port. Counts from 0.
  typedef uint32_t feb_t;                      // FEB number (replacing dcm_link_t)
  
  /// Type for physical pixel on APD.
  typedef uint32_t pixel_t;                    
  
  enum BlockNumerationOffset{
    BLOCK_OFFSET   = 0  ///< do we count blocks from 0 or 1
    ,DIBLOCK_OFFSET = 1  ///< do we count diblocks from 0 or 1
  };
  
  
  // Public Enumerations
  enum DAQValidRange{
    MIN_PIXEL       = 0,
    MAX_PIXEL       = 31,
    MIN_PORT        = 0,
    MAX_PORT        = 63,
    MIN_DCM         = 1,
    MAX_DCM         = 12,
    MIN_DIBLOCK     = 1,
    MAX_DIBLOCK     = 17,
    MIN_DCMCELL     = 0,
    MAX_DCMCELL     = 63,
    MIN_DCMPLANE    = 0,
    MAX_DCMPLANE    = 31,
    MAX_MC_DCMPLANE = 6,   // Max plane for the NDOS muon catcher
    MAX_MC_ND_DCMPLANE = 11   // Max plane for the NearDet muon catcher
  };
  
    /// Shifts and masks for unique online channel
    enum DAQChanMASKS{
      MASK_DETECTOR  = 0xF0000000, //!< Detector ID, (4bit),     valid range 1-4
      MASK_DIBLOCK   = 0x0FC00000, //!< Diblock ID, (6bit),      valid range 1-15
      MASK_DCMID     = 0x003F0000, //!< DCM ID, (6bit),          valid range 1-12
      MASK_DCMPORT   = 0x0000FF00, //!< DCM Link Port (8bit),    valid range 0-63
      MASK_PIXELID   = 0x000000FF, //!< APD Pixel Number (8bit), valid range 0-31

      SHIFT_DETECTOR = 28,
      SHIFT_DIBLOCK  = 22,
      SHIFT_DCMID    = 16,
      SHIFT_DCMPORT  =  8,
      SHIFT_PIXELID  =  0
    };

    /// Shifts and masks for unique offline channel
    enum OfflineChanMASKS{
      MASK_OFFLINE_DETECTOR  = 0xF0000000, //!< Detector ID, (4bit),  valid range 1-4
      MASK_OFFLINE_PLANE     = 0x0FFFFC00, //!< Plane Number,(18bit), valid range 0-930 (far det)
      MASK_OFFLINE_CELL      = 0x000003FF, //!< Cell Number, (10bit), valid range 1-384 (far det)

      SHIFT_OFFLINE_DETECTOR = 28,
      SHIFT_OFFLINE_PLANE    = 10,
      SHIFT_OFFLINE_CELL     = 0,
    };

    /// Types of Block
    enum DetBlock_TYPE {
      MIN_BLOCK_TYPE = 1,
      A_BLOCK        = 1,       //!< Identifier for an "A" type detector block (starts with X measuring plane)
      B_BLOCK    ,		//!< Identifier for a  "B" type detector block (starts with Y measuring plane)
      AFULL_BLOCK,		//!< Identifier for an "AFULL" type detector block (starts with X measuring plane). 32 planes block
      M_BLOCK,                  //!< NDOS Muon catcher Block
      MC_ND_BLOCK,              //!< NearDet Muon catcher Block
      MAX_BLOCK_TYPE=MC_ND_BLOCK
    };

    /// Types of Diblock
    enum DiBlock_TYPE {
      MIN_DIBLOCK_TYPE  = 1,	      //!< Identifier for diblocks of configuration A-B
      AB_DIBLOCK        = 1,	      //!< Identifier for diblocks of configuration A-B
      BA_DIBLOCK        ,	      //!< Identifier for diblocks of configuration B-A
      AA_DIBLOCK        ,	      //!< Identifier for diblocks of configuration A-A
      AFULLAFULL_DIBLOCK,             //!< Identifier for diblocks using a 32/32 configuration
      MC_DIBLOCK,                     //!< Identifider for NDOS Muon catcher
      MC_ND_DIBLOCK,                  //!< Identifider for NearDet Muon catcher
      MAX_DIBLOCK_TYPE = MC_ND_DIBLOCK
    };

    /// Types of Detector View
    enum DetView_TYPE {
      INIT = 0,
      X_VIEW = 1,			//!< Identifier for the X measuring view of the detector (top)
      Y_VIEW		        	//!< Identifier for the Y measuring view of the detector (side)
    };

    /// Types of Detector Plane
    enum DetPlane_TYPE {
      XZ_PLANE = 1,//!< Identifier for the horizontal plane
      YZ_PLANE	   //!< Identifier for the vertical plane
    };

    /// Size of the detectors in blocks
    enum DetBlockSizes {
      DetSize_IPND                = 4,	//!< Size of the IP/ND detector in blocks (4)
      DetSize_NDOS                = 8,	//!< Size of the NDOS detector in blocks (6)
      DetSize_NEARDET             = 8,	//!< Size of the Near detector in blocks (6)
      DetSize_FARDET              = 30,	//!< Size of the Far detector in blocks (30)
      DetSize_FARDET_16K          = 32,	//!< Size of the Far detector in blocks (32)
      DetSize_FARDET_17K          = 32,	//!< Size of the Far detector in blocks (32)
      DetSize_TESTBEAM_2Block_1x1 = 2,
      DetSize_TESTBEAM_3Block_1x1 = 3,
      DetSize_TESTBEAM_2Block_2x2 = 2,
      DetSize_TESTBEAM_3Block_2x2 = 3
    };

    enum DetDiBlockSizes {
       NUM_DIBLOCK_IPND                = 2	//!< Size of the IP/ND detector in diblocks
      ,NUM_DIBLOCK_NDOS                = 4	//!< Size of the NDOS detector in diblocks including muon catcher
      ,NUM_DIBLOCK_NEARDET             = 4	//!< Size of the Near detector in diblocks including muon catcher
      ,NUM_DIBLOCK_FARDET              = 15	//!< Size of the Far  detector in diblocks
      ,NUM_DIBLOCK_FARDET_16K          = 16	//!< Size of the Far  detector in diblocks
      ,NUM_DIBLOCK_FARDET_17K          = 17	//!< Size of the Far  detector in diblocks
      ,NUM_DIBLOCK_TESTBEAM_2Block_1x1 = 1
      ,NUM_DIBLOCK_TESTBEAM_3Block_1x1 = 2
      ,NUM_DIBLOCK_TESTBEAM_2Block_2x2 = 1
      ,NUM_DIBLOCK_TESTBEAM_3Block_2x2 = 2
    };

    enum ModuleParameters {
      NUM_DCM_PORTS            = 64,
      NUM_PIXELS               = 32,
      NUM_CELLS                = 32,
      NUM_CELLS_PER_DCM        = 64,

      NUM_XZ_PLANES_AFULL_BLOCK  = 16,
      NUM_XZ_PLANES_A_BLOCK      = 16,
      NUM_XZ_PLANES_B_BLOCK      = 15,
      NUM_XZ_PLANES_MC_BLOCK     = 6,
      NUM_XZ_PLANES_MC_ND_BLOCK  = 11,

      NUM_YZ_PLANES_AFULL_BLOCK  = 16,
      NUM_YZ_PLANES_A_BLOCK      = 15,
      NUM_YZ_PLANES_B_BLOCK      = 16,

      NUM_YZ_PLANES_MC_BLOCK     = 7,
      NUM_YZ_PLANES_MC_ND_BLOCK  = 11,

      NUM_X_MODULES_IPND                = 2,
      NUM_Y_MODULES_IPND                = 3,
      NUM_X_MODULES_NDOS                = 2,
      NUM_Y_MODULES_NDOS                = 3,
      NUM_X_MODULES_NEARDET             = 3,
      NUM_Y_MODULES_NEARDET             = 3,
      NUM_X_MODULES_FARDET              = 12,
      NUM_Y_MODULES_FARDET              = 12,
      NUM_X_MODULES_TESTBEAM_2Block_1x1 = 1,
      NUM_Y_MODULES_TESTBEAM_2Block_1x1 = 1,
      NUM_X_MODULES_TESTBEAM_3Block_1x1 = 1,
      NUM_Y_MODULES_TESTBEAM_3Block_1x1 = 1,
      NUM_X_MODULES_TESTBEAM_2Block_2x2 = 2,
      NUM_Y_MODULES_TESTBEAM_2Block_2x2 = 2,
      NUM_X_MODULES_TESTBEAM_3Block_2x2 = 2,
      NUM_Y_MODULES_TESTBEAM_3Block_2x2 = 2,

      NUM_DCMS_DIBLOCK_IPND                = 3,
      NUM_DCMS_DIBLOCK_NDOS                = 3,
      NUM_DCMS_DIBLOCK_NEARDET             = 4,
      NUM_DCMS_DIBLOCK_FARDET              = 12,
      NUM_DCMS_MUONCATCHER                 = 2,
      NUM_DCMS_DIBLOCK_TESTBEAM_2Block_1x1 = 2,
      NUM_DCMS_DIBLOCK_TESTBEAM_3Block_1x1 = 2,
      NUM_DCMS_DIBLOCK_TESTBEAM_2Block_2x2 = 2,
      NUM_DCMS_DIBLOCK_TESTBEAM_3Block_2x2 = 2,

      NUM_X_DCMS_IPND                = 1,
      NUM_X_DCMS_NDOS                = 1,
      NUM_X_DCMS_NEARDET             = 2,
      NUM_X_DCMS_FARDET              = 6,
      NUM_X_DCMS_TESTBEAM_2Block_1x1 = 1,
      NUM_X_DCMS_TESTBEAM_3Block_1x1 = 1,
      NUM_X_DCMS_TESTBEAM_2Block_2x2 = 1,
      NUM_X_DCMS_TESTBEAM_3Block_2x2 = 1,

      NUM_Y_DCMS_IPND                = 2,
      NUM_Y_DCMS_NDOS                = 2,
      NUM_Y_DCMS_NEARDET             = 2,
      NUM_Y_DCMS_FARDET              = 6,
      NUM_Y_DCMS_TESTBEAM_2Block_1x1 = 1,
      NUM_Y_DCMS_TESTBEAM_3Block_1x1 = 1,
      NUM_Y_DCMS_TESTBEAM_2Block_2x2 = 1,
      NUM_Y_DCMS_TESTBEAM_3Block_2x2 = 1,

      NUM_CHAN_X_PLANE_IPND                 = 64,
      NUM_CHAN_X_PLANE_NDOS                 = 64,
      NUM_CHAN_X_PLANE_NEARDET              = 96,
      NUM_CHAN_X_PLANE_FARDET               = 384,
      NUM_CHAN_X_PLANE_TESTBEAM_2Block_1x1  = 32,
      NUM_CHAN_X_PLANE_TESTBEAM_3Block_1x1  = 32,
      NUM_CHAN_X_PLANE_TESTBEAM_2Block_2x2  = 64,
      NUM_CHAN_X_PLANE_TESTBEAM_3Block_2x2  = 64,

      NUM_CHAN_Y_PLANE_IPND                 = 96,
      NUM_CHAN_Y_PLANE_NDOS                 = 96,
      NUM_CHAN_Y_PLANE_NEARDET              = 96,
      NUM_CHAN_Y_PLANE_FARDET               = 384,
      NUM_CHAN_Y_PLANE_TESTBEAM_2Block_1x1  = 32,
      NUM_CHAN_Y_PLANE_TESTBEAM_3Block_1x1  = 32,
      NUM_CHAN_Y_PLANE_TESTBEAM_2Block_2x2  = 64,
      NUM_CHAN_Y_PLANE_TESTBEAM_3Block_2x2  = 64,

      NULL_TERM = 0x0,
      TERM_CHAR = 0xFDFCFBFA
    };

}// end of namespace daqchannelmap

#endif	/* _DAQCHANNELMAPCONSTANTS_H */

