#include "DAQDataFormats/RawMicroSlice.h"
#include "DAQDataFormats/DataFormatException.h"

#include <boost/lexical_cast.hpp>
#include <assert.h>
namespace daqdataformats{

//------------------------------------------------------------------------------
// Default constructor
RawMicroSlice::RawMicroSlice(const uint32_t size, const bool resize_buffer, const version_t version):
     RawDAQData( 0
                , (const init_t)           &RawMicroSlice::init
                , (const sizeofdata_t)     &RawMicroSlice::getPredeterminedSize
                , (const setBufferSource_t)&RawMicroSlice::setBufferSource
                , (const print_t)          &RawMicroSlice::printStandard
                , (const readData_t)       &RawMicroSlice::readDataFromBuffer
                )
        , _Header                 (NULL)
        , _TimingMarker           (NULL)
        , _FloatingNanoSlice      (NULL)
        , _HitProbability         (0.0)
        , _nanosliceVersion       (rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION)
        , _microSliceHeaderVersion(version)
{

    /// Check if the size is legal
    if (size < getPredeterminedSize()) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(getPredeterminedSize()) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
    }// end of checking if the size is legal

    /// Set the size of the Internal Buffer
    if(resize_buffer){
        resizeInternalBuffer(size);
    }
    else
        reserveInternalBuffer(size);

    defaultConstructorFloatingSize();
}
//------------------------------------------------------------------------------
RawMicroSlice::RawMicroSlice(const RawMicroSlice& copy_in):
          RawDAQData(copy_in)
        , _Header                 (NULL)
        , _TimingMarker           (NULL)
        , _FloatingNanoSlice      (NULL)
        , _HitProbability         (copy_in._HitProbability)
        , _nanosliceVersion       (copy_in._nanosliceVersion)
        , _microSliceHeaderVersion(copy_in._microSliceHeaderVersion)
{
    lastOperationsInDeepCopy(copy_in);
}

//------------------------------------------------------------------------------
RawMicroSlice::RawMicroSlice(const void* microslice_buffer, const uint32_t size, const version_t microsliceheader_version):
     RawDAQData( 0
                , (const init_t)           &RawMicroSlice::init
                , (const sizeofdata_t)     &RawMicroSlice::getPredeterminedSize
                , (const setBufferSource_t)&RawMicroSlice::setBufferSource
                , (const print_t)          &RawMicroSlice::printStandard
                , (const readData_t)       &RawMicroSlice::readDataFromBuffer
                )
        , _Header                 (NULL)
        , _TimingMarker           (NULL)
        , _FloatingNanoSlice      (NULL)
        , _HitProbability         (0.0)
        , _nanosliceVersion       (rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION)
        , _microSliceHeaderVersion(microsliceheader_version) ///< we'll figure it out in setBufferSource
{
        /// Check if the size is legal
        if (size < getPredeterminedSize()) {
            std::string msg("An external input buffer was given to the ");
            msg.append("RawMicroSlice constructor that was too small (size=");
            msg.append(boost::lexical_cast<std::string>(size) + ", minimum size=");
            msg.append(boost::lexical_cast<std::string>(getPredeterminedSize()) + ").");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
        }// end of checking if the size is legal

        /// In addition to standard shallow copy routine, set the maximum size
        reserveInternalBuffer(size);

        RawDAQData::copy(microslice_buffer);
}


//------------------------------------------------------------------------------
// Destructor
RawMicroSlice::~RawMicroSlice() {
    if(_FloatingNanoSlice) delete _FloatingNanoSlice;
    if(_Header)            delete _Header;
    if(_TimingMarker)      delete _TimingMarker;
}

//------------------------------------------------------------------------------
bool RawMicroSlice::setVersion(const version_t version)
{
    if(_microSliceHeaderVersion != rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION){
            std::string msg("Cannot set MicroSlice to version " + boost::lexical_cast<std::string>(version));
            msg.append("\n It already has been set to " +  boost::lexical_cast<std::string>(_microSliceHeaderVersion));
            GENERATE_DATAFORMAT_EXCEPTION(msg);
            return false;
    }

    _microSliceHeaderVersion = version;

    //if(_Buffer){
    //    setBufferSource(_Buffer);
    //}

    if(_Header){
        _Header->setVersionSoft(_microSliceHeaderVersion);
    }

    return true;
}


//------------------------------------------------------------------------------
// Initialize a new MicroSlice (or reset it)
void RawMicroSlice::init()
{

    /// This is needed for the formats with variable size
    /// If buffer is internal, need to resize it
    if(_shouldBufferBeInternal){
        resizeInternalBuffer(rawmicro::PREDETERMINED_MICROSLICE_SIZE);
    }

    // Initialize the header and timing marker
    if(_microSliceHeaderVersion == rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION)
        _Header->initSoft();
    else
        _Header->init();

    _TimingMarker->init();
}

//------------------------------------------------------------------------------
// Method to Add a packed Nanoslice to the _Buffer
// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
// and adds an appropriate flag to the MicroSlice Header.
void RawMicroSlice::addNanoSlice(const void* buffer, const bool check_mc_flag){

  /// Construct a temporary nanoslice with external buffer
  const RawNanoSlice  tmp_nanoslice(buffer);
  /// This is its size
  const uint32_t      nanoslice_size = tmp_nanoslice.sizeofdata();

  /// Now add that much data to MicroSlice
  addData(buffer, nanoslice_size, check_mc_flag);
}

//------------------------------------------------------------------------------
// Method to Add data to the _Buffer
// If check_mc_flag = true, the method checks whether the nanoslice is MC generated
// and adds an appropriate flag to the MicroSlice Header.
void RawMicroSlice::addData(const void* buffer, const uint32_t nwords, const bool check_mc_flag){

    /// Add the data from buffer into the _Buffer
    readDataGeneral(buffer, nwords);

    /// After Reading data, Buffer could have changed position
    checkBufferInternalness();

    /// Add byte count to the MicroSlice Header
    _Header->addByteCount(nwords * sizeof(uint32_t));

    /// Check if the NanoSlice is the MC generated
    /// If it is, set the appropriate MicroSlice MC flag
    if(check_mc_flag){
        RawNanoSliceHeader nano_header(buffer, rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION);
        const uint32_t FEBStatus = nano_header.getFEBStatus();
        if(FEBStatus > 0x0000003F) setMonteCarloFlag(true);
    }// end of checking mc flag

}// end of addData


//------------------------------------------------------------------------------
// Method to generate a Microslice from scratch
void RawMicroSlice::generate(const version_t version_of_nanoslice){

    _nanosliceVersion = (version_t)version_of_nanoslice;

    /// Set the MC flag
    setMonteCarloFlag(true);

    /// Generate the timing marker of the microslice
    _TimingMarker->generate();

    // Loop over all the front end boards generating hits
    // and writing them to the buffers
    for(uint32_t boardID = 0; boardID < rawmicro::MAXIMUM_FRONTEND_BOARDS; ++boardID){

        /// Declare the front end boards simulators
        FEBSimulator FrontEndBoard(boardID, _HitProbability, _nanosliceVersion);

        for(int pixelID = 0; pixelID < 32; ++pixelID){
	  /// Attempt to generate a nanoslice hit on a given pixel
	  FrontEndBoard.generateNanoSlice(pixelID);
	  
	  /// Check to see if it has nonzero data
	  if(FrontEndBoard.nonEmptyNanoSlice()){
	    // /Add the nanoslice from the FEB to the buffer
	    addNanoSlice(FrontEndBoard.getLastNanoSlice()->getBuffer());
	  } // endif NonEmptyNanoSlice
	} // endfor pixelID
    } // endfor boardID
}// end of RawMicroSlice::unpack()


//------------------------------------------------------------------------------
// Getters
daqdataformats::RawMicroSliceHeader*  RawMicroSlice::getHeader()           const  {
    return _Header;
}                    // Get MicroSlice Header pointer
daqdataformats::RawTimingMarker*      RawMicroSlice::getTimingMarker()     const  { return _TimingMarker; }              // Get Timing Marker pointer
daqdataformats::RawNanoSlice*         RawMicroSlice::getFloatingNanoSlice()const  { return _FloatingNanoSlice;}
bool                  RawMicroSlice::getDataPresent()    const
{
    return _Header->getDataPresent();
}    // Get Method for Data present bit
bool                  RawMicroSlice::getMonteCarloFlag() const
{ 
    return _Header->getMonteCarloFlag();
} // Get Method for SIM bit
uint32_t              RawMicroSlice::getByteCount()      const
{
    return _Header->getByteCount();
}      // Get Method for Byte Count

uint32_t              RawMicroSlice::getLowWord()        const { return _TimingMarker->getLowWord();}  // Get method for lower half of timing word
uint32_t              RawMicroSlice::getHighWord()       const { return _TimingMarker->getHighWord();} // Get method for upper half of timing word
uint64_t              RawMicroSlice::getTime()           const { return _TimingMarker->getTime();}     // Get the time-stamp as a single 64-bit value */
float                 RawMicroSlice::getHitProbability() const { return _HitProbability;}

//------------------------------------------------------------------------------
uint32_t RawMicroSlice::getNumNanoSlices() const { 

  /// Could be zero nanoslices
  if(!areThereNanoSlices())return 0;

  const uint32_t nanoslice_size = getNanoSliceSize();
  if(nanoslice_size == 0){
            std::string msg("From RawMicroSlice::getNumNanoSlices. NanoSLice size is zero.");
            GENERATE_DATAFORMAT_EXCEPTION(msg);
  }// end of checking whether nanoslice size is zero

  return (sizeofdata() - rawmicro::PREDETERMINED_MICROSLICE_SIZE) / nanoslice_size;
}

//------------------------------------------------------------------------------
// Sometimes we don't know the NanoSlice size yet,
// But we need to know if there are NanoSlices at all
bool RawMicroSlice::areThereNanoSlices() const { return (sizeofdata() - rawmicro::PREDETERMINED_MICROSLICE_SIZE > 0);}

//------------------------------------------------------------------------------
//Setters
bool RawMicroSlice::setMonteCarloFlag (const bool mc_flag)
{ 
    return _Header->setMonteCarloFlag(mc_flag);
}
bool RawMicroSlice::setLowWord        (const uint32_t word)      { return _TimingMarker->setLowWord (word);}
bool RawMicroSlice::setHighWord       (const uint32_t word)      { return _TimingMarker->setHighWord(word);}
bool RawMicroSlice::setTime           (const uint64_t timestamp) { return _TimingMarker->setTime(timestamp);}
bool RawMicroSlice::setHitProbability (const float hit_prob)     { _HitProbability = hit_prob; return true; }

//------------------------------------------------------------------------------
uint32_t RawMicroSlice::getNanoSliceSize() const {

    /// Could be zero nanoslices, then we don't know the nanoslice size
    if(!areThereNanoSlices())return 0;

    /// If floating nanoslice is not set, set it to zeroth nanoslice
    if(_FloatingNanoSlice == NULL && !setFloatingNanoSlice(0)) return 0;

    return _FloatingNanoSlice->sizeofdata();
}// end of getNanoSliceSize


//------------------------------------------------------------------------------
bool RawMicroSlice::setFloatingNanoSlice (const uint32_t inano) const {

    /// If there are no nanoslices, then can't the floating NanoSlice
    if(!areThereNanoSlices())return false;

    /// Check whether NanoSlice is valid
    /// If nanoslice is zero, then nothing to check
    if(inano && inano >= getNumNanoSlices()){
        //std::cout<<"Warning from RawMicroSlice::setFloatingNanoSlice : setting floating nanoslice to "<<inano
        //         <<"-th nanoslice, whereas only "<<this->getNumNanoSlices()<<"nanoslices are in the current MicroSlice."<<"\n";
        return false;
    }// end of checking whether NanoSlice is valid

    uint32_t* pointer = &RAWBUFF32[0] + rawmicro::PREDETERMINED_MICROSLICE_SIZE;

    /// counting nanoslices from zero
    /// if inano is zero, then don't need to calculate NanoSlice size
    if(inano) pointer += inano * getNanoSliceSize();

    /// See if we need to create a new Nanoslice or just set the pointer
    if(_FloatingNanoSlice == NULL){
        _FloatingNanoSlice = new RawNanoSlice(pointer, _nanosliceVersion);
        _nanosliceVersion  = _FloatingNanoSlice->getHeader()->getVersion();
    }
    else{
        _FloatingNanoSlice->setBufferSource(pointer);
    }
  
    return true;
}

//------------------------------------------------------------------------------
/// Setting Buffer, Header and Timing marker. Don't bother Floating NanoSlice
bool RawMicroSlice::setBufferSource(const void* pointer){

    _Buffer = (void*)pointer;

    _shouldBufferBeInternal = isInternalBuffer();

    if(!_shouldBufferBeInternal && _microSliceHeaderVersion == rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION){
        _microSliceHeaderVersion = daqdataformats::RawMicroSliceHeader::figureOutTheVersionStatic(pointer);
        //assert(_microSliceHeaderVersion != rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION);
        if(_Header) {
            _Header->setVersionSoft(_microSliceHeaderVersion);
        }
    }

    /// See if header and Timing marker are already constructed
    if(_Header) {
            _Header->setBufferSource(_Buffer);
            const uint32_t* timing_marker_pointer = &RAWBUFF32[0] + _Header->sizeofdata();
            return _TimingMarker->setBufferSource(timing_marker_pointer);
    }

    uint32_t* timing_marker_pointer = &RAWBUFF32[0];

    _Header                = new daqdataformats::RawMicroSliceHeader(_Buffer, _microSliceHeaderVersion);
    timing_marker_pointer += _Header->sizeofdata();

    _TimingMarker          = new RawTimingMarker(timing_marker_pointer);

    _function_sizeofdata   = (sizeofdata_t) &RawMicroSlice::sizeofdata;

    return true;
}// end of setBufferSource1

//------------------------------------------------------------------------------
// Print the MicroSlice structure
void RawMicroSlice::printStandard(std::ostream& os) const {

    const uint32_t size_of_data = sizeofdata();
    os<<"----------MicroSlice----------------------------------------size="<<size_of_data<<std::endl;
    printBuffer(os);
    if(_microSliceHeaderVersion != rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION){
        _Header->print(os);
    }
    
    _TimingMarker->print(os);


    const uint32_t num_nanoslices = getNumNanoSlices();

    os<<std::endl<<"----------NumNanoSlices-------="<<num_nanoslices<<std::endl;

    /// If number is nonzero, need to print them
    if(num_nanoslices){
    /// Get a copy of the MicroSlice this is so that we don't screw up the internal structure
        const RawMicroSlice micro_copy(_Buffer, 3);

        for(uint32_t i = 0; i < num_nanoslices; ++i){
            micro_copy.setFloatingNanoSlice(i);
            micro_copy.getFloatingNanoSlice()->print(os);
        }// end of loop over nanoslices
    }// end of checking whether the number of nanoslices is non-zero


}// end of print1

//------------------------------------------------------------------------------
// MicroSlice _Buffer size in words
uint32_t RawMicroSlice::sizeofdata() const
{
    return _Header->getMicroSliceSize();
}

//------------------------------------------------------------------------------
// This is a general method to read the microslice from microslice_buffer to _Buffer
void* RawMicroSlice::readDataFromBuffer(const void* microslice_buffer)
{
    /// Figure out version
    if(getVersion() == rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION){
        setVersion(RawMicroSliceHeader::figureOutTheVersionStatic(microslice_buffer));
    }

    /// First, read Header and Timing marker
    void* microslice_buffer_new = readDataGeneral(microslice_buffer, rawmicro::PREDETERMINED_MICROSLICE_SIZE, 0);

    /// Now, read the rest of the the _Buffer
    microslice_buffer_new = readDataGeneral(microslice_buffer_new
                                           ,sizeofdata() - rawmicro::PREDETERMINED_MICROSLICE_SIZE
                                           ,rawmicro::PREDETERMINED_MICROSLICE_SIZE
                                           );
    /// Check pointers after each read or resize buffer
    checkBufferInternalness();
    
    return microslice_buffer_new;
}

//------------------------------------------------------------------------------
readDataStatus RawMicroSlice::readData(const int fd)
{

  /// Check if the file descriptor is invalid
  if(fd < 0) return rds_error;

  char* current_buffer = (char*)_Buffer;

  /// Read the Header
  ssize_t bytes_read = read(fd, current_buffer, (size_t)(RAWMICROSLICEHEADER_SIZE*sizeof(uint32_t)));
  current_buffer += bytes_read;

  /// Figure out version
  if(getVersion() == rawdaqdataformat::DAQDATAFORMAT_UNKNOWN_VERSION){
        setVersion(RawMicroSliceHeader::figureOutTheVersionStatic(_Buffer));
  }

  /// Check if the reading was OK
  if (bytes_read == -1 || bytes_read == 0)                   return rds_error;

  /// Check if the reading was complete
  if ((uint32_t)bytes_read != RAWMICROSLICEHEADER_SIZE*sizeof(uint32_t)) return rds_incomplete;
      
  /// After reading the header, now we know the whole MicroSlice size
  /// In words
  const uint32_t microslice_size         = sizeofdata();
  const uint32_t rest_of_microslice_size = microslice_size - RAWMICROSLICEHEADER_SIZE;

  /// Check whether we need to adjust the Internal Buffer size
  if(_shouldBufferBeInternal){
      resizeInternalBuffer(microslice_size);
  }

  /// Read the rest of the MicroSlice
  bytes_read = read(fd, current_buffer, (size_t)rest_of_microslice_size*sizeof(uint32_t));
  
  /// Check if the reading was OK
  if (bytes_read == -1 || bytes_read == 0)                   return rds_error;

  /// Check if the reading was complete
  if ((uint32_t)bytes_read != rest_of_microslice_size*sizeof(uint32_t)) return rds_incomplete;
  
  return rds_complete;
}

}// end of namespace
